#!/usr/bin/env python

### IMPORT MODULES ###
import re
import urwid
from typing import Union

### CUSTOM WIDGETS ###




### CUSTOM URWID EDIT ###
class CommandEdit(urwid.Edit):
    _metaclass_ = urwid.signals.MetaSignals
    signals = ['done']

    def __init__(self, caption=u"", edit_text=u"", multiline=False, align=urwid.LEFT, wrap=urwid.SPACE, allow_tab=False, edit_pos=None, layout=None, mask=None, history_size=50, enable=True):

        self._history = []
        self._history_idx = -1
        self._history_size = history_size
        self._enabled = enable

        # Run standard urwid.Edit init
        urwid.Edit.__init__(self, caption, edit_text, multiline, align, wrap, allow_tab, edit_pos, layout, mask)

    def keypress(self, size, key):

        # Curser Up to go back in history
        if self._command_map[key] == urwid.CURSOR_UP:
            self._history_idx = min(self._history_idx+1, len(self._history)-1)
            self.set_edit_text(self._history[self._history_idx])
            return

        # Cursor Down to go forward in history
        if self._command_map[key] == urwid.CURSOR_DOWN:
            self._history_idx = max(self._history_idx-1, -1)
            if self._history_idx == -1:
                self.set_edit_text('')
            else:
                self.set_edit_text(self._history[self._history_idx])
            return

        # Enter to emit done signal
        elif key == 'enter':
            self._history.insert(0, self.get_edit_text()) 
            if len(self._history) > self._history_size:
                self._history.pop()
            self._history_idx = -1

            urwid.emit_signal(self, 'done', self, self.get_edit_text())
            self.set_edit_text('')
            # super(SerialCommand, self).set_edit_text('')
            return

        # Escape to clear text
        elif key == 'esc':
            self._history_idx = -1
            self.set_edit_text('')
            # super(SerialCommand, self).set_edit_text('')
            return

        urwid.Edit.keypress(self, size, key)

    def selectable(self):
        return self._enabled

    def enable(self, enable):
        self._enabled = enable

### CUSTOM OPTION ###
class UserOption(urwid.WidgetWrap):
    _metaclass_ = urwid.signals.MetaSignals
    signals = ['value_change']

    def __init__(self, option_name : str, value : Union[list, str, int, float], caption : str="", increment : Union[int, float]=None, limits : Union[int, float, list]=[None, None], enter_fires_change_event : bool=True, enabled : bool=True):
        # Variables
        self._option_name = option_name
        self._value_list = None             # holds the list of selectable values, if passed values is a list, else will be none
        self._current_value = 0             # holds the current value, or current index of selected _values_list, if value is a list of selectable options
        self._shown_value = 0               # holds the current value shown to the user
        self._caption = caption
        self._increment = increment
        self._limits = limits               # limits that self._current_value and self._shown value can take, inclusive
        self._enter_fires_change_event = enter_fires_change_event
        self._enabled = enabled
        self._editing = False


        # Set self._value_list, self._current_value and self._shown_value
        if type(value) == list:
            # value contains a list of selectable options
            self._value_list = value
            self._current_value = 0
            self._shown_value = 0
            self._increment = 1                         # to increment list value by 1
            self._limits = [0, len(self._value_list)-1]   # limits are 0 and length of list

            # Initialise urwid selectable icon to contain user option value
            self._option_text = urwid.SelectableIcon("???")
            self._option_text.set_align_mode("center")
            self._set_option_text(self._current_value)
            self._set_option_text_curs_pos()

        else:
            self._current_value = value
            self._shown_value = value

            if type(value) == str:
                self._increment = None # incrementing a string does not make sense

            # Initialise urwid edit text to contain user option value
            self._option_text = urwid.Edit(edit_text="???", align='center', edit_pos=None)
            self._set_option_text(self._current_value)
            self._set_option_text_curs_pos()

        # Initialise remaining urwid widgets
        if caption.strip() != "":
            self._caption_text = urwid.Text(caption, align='center')
            option = urwid.Columns([self._caption_text, self._option_text])
        else:
            option = self._option_text
        option = urwid.AttrMap(option, None, focus_map='reversed')
        self.__super.__init__(option)


    def _set_option_text(self, value):
        if type(self._option_text) == urwid.SelectableIcon:
            txt = "< %s >"%(str(self._value_list[value]))
            self._option_text.set_text(txt)
        elif type(self._option_text) == urwid.Edit:
            if self._increment != None:
                txt = "< %s >"%(str(value))
            else:
                txt = "%s"%(str(value))
            self._option_text.set_edit_text(txt)

    def _set_option_text_curs_pos(self):
        if type(self._option_text) == urwid.SelectableIcon:
            tmp_list = [str(x) for x in self._value_list]
            curs_pos = len(max(tmp_list, key=len)) + 5 # to hide the cursor so don't see it flashing, +4 for < and > and two spaces then +1 to put at end
            self._option_text._cursor_position = curs_pos
        elif type(self._option_text) == urwid.Edit:
            # doesn't seem to work at the moment
            curs_pos = len(self._option_text.get_edit_text()) - 2 # to move cursor to last place, doesn't seem to be working
            self._option_text.edit_pos = curs_pos

    
    def sizing(self):
        return frozenset([urwid.FLOW])

    def set_option_name(self, option_name : str):
        self._option_name = option_name

    def get_option_name(self):
        return self._option_name

    def set_caption(self, caption : str):
        self._caption = caption
        self._caption_text.set_text(caption)

    def get_caption(self):
        return self._caption

    def set_value(self, value : Union[str, int, float]):
        if type(self._option_text) == urwid.SelectableIcon:
            if type(value) == int or type(value) == float:
                if type(value) == float and not value.is_integer():
                    raise RuntimeError('The passed list option index be an integer not a float.')
                if value < 0 or value > self._limits[1]:
                    raise RuntimeError('The passed list option index is outside the selectable option value list range.')
                self._current_value = int(value)
            elif type(value) == str:
                if value in self._value_list:
                    self._current_value = self._value_list.index(value)

        elif type(self._option_text) == urwid.Edit:            
            if self._limits[0] != None:
                value = max(value, self._limits[0])
            if self._limits[1] != None:
                value = min(value, self._limits[1])
            self._current_value = value

        # set option text
        self._shown_value = self._current_value
        self._set_option_text(self._current_value)

    def get_value(self):
        if type(self._option_text) == urwid.SelectableIcon:
            return self._value_list[self._current_value]
        elif type(self._option_text) == urwid.Edit:
            # attempt to conver to str, int or float
            if re.match('^[0-9]*$', self._current_value): # integer
                return int(self._current_value)
            elif re.match('^[0-9\.]*$', self._current_value): # float
                return float(self._current_value)
            elif self._current_value.lower() == 'true': # true
                return True
            elif self._current_value.lower() == 'false': # false
                return False

            return str(self._current_value) # else return as string

    def get_value_index(self):
        if type(self._option_text) == urwid.SelectableIcon:
            return self._current_value
        elif type(self._option_text) == urwid.Edit:
            return None

    def set_option_list(self, values_list : list, index : int=0):
        if type(self._option_text) == urwid.SelectableIcon:
            self._value_list = values_list
            self._limits = [0, len(self._value_list)-1]   # limits are 0 and length of list
            self.set_value(idx)
            self._set_option_text(self._current_value)
            self._set_option_text_curs_pos()
            
            return True
        else:
            return False

    def get_option_list(self):
        return self._value_list

    def selectable(self):
        return self._enabled

    def enable(self, enable):
        self._enabled = enable

    def valid_numerical_char(self, ch):
        """
        Return true for decimal digits.
        """
        return len(ch)==1 and ch in "0123456789.-"

    def valid_char(self, ch):
        """
        Filter for text that may be entered into this widget by the user

        :param ch: character to be inserted
        :type ch: bytes or unicode

        This implementation returns True for all printable characters.
        """
        return urwid.util.is_wide_char(ch,0) or (len(ch)==1 and ord(ch) >= 32)

    def mouse_event(self, size, event, button, x, y, focus):
        pass 
        # if button != 1 or not urwid.util.is_mouse_press(event):
        #     return False

        # return True
        

    def keypress(self, size, key):
        # Handle left/right arrow keypresses
        if self._command_map[key] == urwid.CURSOR_LEFT or self._command_map[key] == urwid.CURSOR_RIGHT:
            return self._handle_left_right_keypresses(size, key)

        # Handle up/down arrow keypresses
        elif self._command_map[key] == urwid.CURSOR_UP or self._command_map[key] == urwid.CURSOR_DOWN:
            return self._handle_up_down_keypresses(size, key)

        # Handle enter keypress
        elif key == 'enter':
            return self._handle_enter_keypress(size, key)

        # Handle escape keypress
        elif key == 'esc':
            self._editing = False
            self.set_value(self._current_value)
            return # nothing to show keypress has been handled

        # Handle printable character keypress or backspace or delete
        if self.valid_char(key) or key == 'backspace' or key == 'delete':
            return self._handle_edit_keypress(size, key)
        
        # So higher widgets can deal with keypress
        return key


    def _handle_left_right_keypresses(self, size, key):
        # If not currently being edited and increment is not none, increment/decrement value
        if self._editing == False and self._increment != None:
            if self._command_map[key] == urwid.CURSOR_LEFT:
                self._decrement_value()
            else:
                self._increment_value()

            # show new value
            self._set_option_text(self._shown_value)

            # fire change event if self._enter_fires_change_event is false, and only fire change event if current and shown is different
            if self._enter_fires_change_event == False and (self._current_value != self._shown_value):
                self.set_value(self._shown_value)
                self._emit('value_change')

        elif self._editing or self._increment == None:
            p = self._option_text.edit_pos

            if self._command_map[key] == urwid.CURSOR_LEFT:
                if p == 0: 
                    return key
                p = urwid.util.move_prev_char(self._option_text.get_edit_text(), 0, p)
                self._option_text.set_edit_pos(p)

            elif self._command_map[key] == urwid.CURSOR_RIGHT:
                if p >= len(self._option_text.get_edit_text()): 
                    return key
                p = urwid.util.move_next_char(self._option_text.get_edit_text(), p, len(self._option_text.get_edit_text()))
                self._option_text.set_edit_pos(p)

    def _handle_up_down_keypresses(self, size, key):
        self._editing = False

        if self._shown_value != self._current_value:
            if self._enter_fires_change_event:
                # enter cause event change, reset to current value as enter was not pressed
                self.set_value(self._current_value)
            else:
                self.set_value(self._shown_value)
                self._emit('value_change')
        
        return key

    def _handle_enter_keypress(self, size, key):
        if self._current_value != self._shown_value and (self._enter_fires_change_event or self._editing):
            self._editing = False
            self.set_value(self._shown_value)
            self._emit('value_change')

    def _handle_edit_keypress(self, size, key):
        # If valid printable character, or backspace or delete and increment is none, assume editing a string
        if (self.valid_char(key) or key == 'backspace' or key == 'delete') and self._increment == None:
            # handle key
            self._editing = True
            urwid.Edit.keypress(self._option_text, size, key)
            self._shown_value = self._option_text.get_edit_text()

        # Else if valid numerical character, or backspace or delete and increment is not none, assume editing a number
        elif (self.valid_numerical_char(key) or key == 'backspace' or key == 'delete') and self._increment != None:
            # handle key
            self._editing = True
            current_text = self._option_text.get_edit_text()
            urwid.Edit.keypress(self._option_text, size, key)
            tmp_txt = self._option_text.get_edit_text()

            # don't allow editing of starting '< ' and ending ' >' substrings
            if self._increment != None and (tmp_txt[0:2] != '< ' or tmp_txt[-2:] != ' >'):
                self._option_text.set_edit_text(current_text)
                return
            
            if self._increment != None:
                tmp_txt = tmp_txt[2:-2] # ignore starting and ending substrings
            
            if not (tmp_txt.strip() == '.' or tmp_txt.strip() == '-' or tmp_txt.strip() == ''):
                try:
                    tmp_val = float(tmp_txt) 
                    self._shown_value = tmp_val
                except ValueError:
                    # reset back to shown value
                    self._option_text.set_edit_text(current_text)
            else:
                self._shown_value = 0
                self._option_text.set_edit_text('< 0 >')

    def _increment_value(self):
        self._shown_value += self._increment
        if self._limits[1] != None and self._shown_value > self._limits[1]:
            if type(self._option_text) == urwid.SelectableIcon:
                self._shown_value = self._limits[0]     # wrap value
            elif type(self._option_text) == urwid.Edit:
                self._shown_value = self._limits[1]     # do not wrap value

    def _decrement_value(self):
        self._shown_value -= self._increment
        if self._limits[0] != None and self._shown_value < self._limits[0]:
            if type(self._option_text) == urwid.SelectableIcon:
                self._shown_value = self._limits[1]     # wrap value
            elif type(self._option_text) == urwid.Edit:
                self._shown_value = self._limits[0]     # do not wrap value


### CUSTOM URWID FRAME ###
class CustomFrame(urwid.Frame):

    def keypress(self, size, key):

        if key == 'tab':
            if self.focus_position == 'header' and self.body.selectable():
                self.focus_position = 'body'
            elif self.focus_position == 'header' and self.footer.selectable():
                self.focus_position = 'footer'

            elif self.focus_position == 'body' and self.footer.selectable():
                self.focus_position = 'footer'
            elif self.focus_position == 'body' and self.header.selectable():
                self.focus_position = 'header'

            elif self.focus_position == 'footer' and self.header.selectable():
                self.focus_position = 'header'
            elif self.focus_position == 'footer' and self.body.selectable():
                self.focus_position = 'body'

        urwid.Frame.keypress(self, size, key)


### CUSTOM URWID TEXT ###
class CustomText(urwid.Text):
    _metaclass_ = urwid.signals.MetaSignals
    signals = ['resize']

    _size = (0,0)
    _first_render = False
    
    def render (self, size, focus = False):
        canvas = super(CustomText, self).render(size, focus)
        new_size = (canvas.rows(), canvas.cols())
        # check if to raise *on_resize* event
        if new_size != self._size and not self._first_render:
            self._emit('resize', self._size, new_size)
        self._size = new_size
        self._first_render = False
        return canvas

    def get_size(self):
        return self._size


### CUSTOM URWID FILLER ###
class CustomFiller(urwid.Filler):
    _metaclass_ = urwid.signals.MetaSignals
    signals = ['resize']

    def __init__(self, body, valign=urwid.MIDDLE, height=urwid.PACK, min_height=None, top=0, bottom=0):
        self._size = (0,0)
        self._first_render = False

        urwid.Filler.__init__(self, body, valign, height, min_height, top, bottom)
    
    def render (self, size, focus = False):
        canvas = super(CustomFiller, self).render(size, focus)
        new_size = (canvas.rows(), canvas.cols())
        # check if to raise *on_resize* event
        if new_size != self._size and not self._first_render:
            self._emit('resize', self._size, new_size)
        self._size = new_size
        self._first_render = False
        return canvas

    def get_size (self):
        return self._size


#!/usr/bin/env python

# ### IMPORT MODULES ###
import urwid
from typing import Callable, Union
from TerminalUI.CustomUrwidWidgets import *


### TERMINAL UI CLASS ###
class TerminalUI():
    """The TerminalUI class is used to create a terminal user interface that can be used for asynchronous read/write operations (e.g. Serial Communications) with an optional options menu for changing application code. All that is required is a user-defined function to be called when a write command is entered, and if desired an options menu and a user-defined function to be called when an option value is changed.

    Raises:
        RuntimeError: on construction, if the options is not a dictionary
        RuntimeError: on construction, if the value for each key in options is not a dictionary or a tuple
        RuntimeError: on construction, if the length of the tuple for the group name key value has a length greater than 3
        RuntimeError: on construction, if all options names are not unique
        RuntimeError: on construction, if it is a selectable option list passed as part of a tuple and the tuple has a length greater than 4
        RuntimeError: on construction, if it is an editable value option passed as part of a tuple and the tuple has a length greater than 6
        RuntimeError: on construction, if the value for the option name key is not a str, int, float, list or tuple
    """

    ### INITIALISE ###
    def __init__(self, title : str, command_entered_callback : Callable[['TerminalUI', str], None], options : dict=None, option_item_selected_callback : Callable[['TerminalUI', str, Union[int, float, bool, str], int], None]=None, options_width_weight : float=0.3):
        """The constructor for the TerminalUI class, will initiliase variables and screen widgets. 
        See https://github.com/jmount1992/TerminalUI/tree/main/examples/option_example.py for details and a code example on setting the options argument

        Args:
            title (str): The title wishing to place at the top of the TerminalUI screen
            command_entered_callback (Callable[[TerminalUI, str], None]): The function to run when a command is entered by the user
            options (dict, optional): A dictionary containing options and their values as well as display information. Defaults to None. 
            option_item_selected_callback (Callable[[TerminalUI, str, Union[int, float, bool, str], int], None], optional): The function to run when an option change event fires. Defaults to None. 
            options_width_weight (float, optional): The amount of the screen width to use for the options area, rest will be used for receive textbox. Defaults to 0.3.

        Raises:
            RuntimeError: if the options is not a dictionary
            RuntimeError: if the value for each key in options is not a dictionary or a tuple
            RuntimeError: if the length of the tuple for the group name key value has a length greater than 3
            RuntimeError: if all options names are not unique
            RuntimeError: if it is a selectable option list passed as part of a tuple and the tuple has a length greater than 4
            RuntimeError: if it is an editable value option passed as part of a tuple and the tuple has a length greater than 6
            RuntimeError: if the value for the option name key is not a str, int, float, list or tuple
        """


        # CONSTANTS
        self.OPTIONS_WIDTH_WEIGHT = options_width_weight

        # VARIABLES
        self._command_entered_callback = command_entered_callback
        self._option_item_selected_callback = option_item_selected_callback
        self._command_debug_visible = True

        # UI/WIDGET SETUP
        # Receive Area
        self._initialise_receive_area()

        # Options Area
        if not (options == None or options == {}):
            self._initialise_options_area(options)
        
        # Body Area - Receive and Options
        if not (options == None or options == {}):
            body = urwid.Columns([('weight', 1-self.OPTIONS_WIDTH_WEIGHT, self.receive_area), ('weight', self.OPTIONS_WIDTH_WEIGHT, self.options_area)])
        else:
            body = self.receive_area

        # Command Area
        self._initialise_command_area()

        # Main Frame
        self.title_text = urwid.Text('%s'%(title), align='center')
        self.main_frame = CustomFrame(body, header=self.title_text, footer=self.command_area)

        if options == None or options == {}:
            self.main_frame.focus_position = 'footer'


    ### RUN ###
    def run(self, redraw_period=0.1):
        """Will start running the TerminalUI, drawing widgets to screen and capturing events.

        Args:
            redraw_period (float, optional): Will redraw the screen every X seconds. Used to update widgets when they are updated from a separate thread. If set to 0, no update will occur. Defaults to 0.1.
        """

        self.main_loop = urwid.MainLoop(self.main_frame, palette=[('reversed', 'standout', '')])
        if redraw_period != 0:
            self.main_loop.set_alarm_in(redraw_period, self._redraw, user_data=redraw_period)
            self.main_loop.run()

    def exit(self):
        """Exits the urwid event loop. Use this to close the TerminalUI after calling TerminalUI.run().
        """
        raise urwid.ExitMainLoop
    
    def _redraw(self, main_loop, redraw_period=0.1):
        """Redraws the screen and set an alarm to redraw in X seconds.

        Args:
            main_loop ([type]): the urwid MainLoop object
            redraw_period (float, optional): The number of seconds until redraw the screen again. Set to 0 to not redraw. Defaults to 0.1.
        """
        main_loop.draw_screen()
        if redraw_period != 0:
            main_loop.set_alarm_in(redraw_period, self._redraw, redraw_period)

    ### COMMAND FUNCTIONS ###
    def _initialise_command_area(self):
        """Initiliases the command area which contains Edit and Text widgets."""

        # Setup Command Area Widgets
        self.command_edit = CommandEdit(">>")
        self.command_txt = urwid.Text('Debug')
        self.command_area = self._get_command_area()

        # Connect Signals
        urwid.connect_signal(self.command_edit, 'done', self._on_command_enter) 

    def _on_command_enter(self, widget : urwid.Widget, command : str):
        """Event handler that calls the used defined command_entered_callback function passed to the constructor.

        Args:
            widget (urwid.Widget): the widget that signalled the event.
            command (str): the string entered into the edit widget
        """

        self._command_entered_callback(self, command)

    def _get_command_area(self) -> urwid.Widget:
        """Builds and returns the complete widget for the command area.

        Returns:
            urwid.Widget: returns the widget for the command area
        """
        command_pile = urwid.Pile([self.command_edit, urwid.Divider('-'), self.command_txt])
        if self._command_debug_visible:
            command_area = urwid.LineBox(command_pile)
        else:
            command_area = urwid.LineBox(self.command_edit)
        
        return command_area

    def set_command_debug_text(self, text, clear : bool=True):
        """Sets the text within the command debug textbox.
        
        Args:
            clear (bool, optional): used to specify if wish to clear current text within the textbox. Defaults to True.
        """
        if clear or self.command_txt.get_text()[0].strip() == '':
            self.command_txt.set_text(text)
        else:
            current_txt = self.command_txt.get_text()[0]
            self.command_txt.set_text(current_txt + text)

    def command_debug_text_visible(self, visible):
        """Used to show/hide the command debug textbox."""

        self._command_debug_visible = visible
        self.main_frame.footer = self._get_command_area()

    def enable_command(self, enable : bool):
        """Used to enable/disable the command edit text area.

        Args:
            enable (bool): used to specify if the write command edit text is selectable
        """

        self.command_edit.enable(enable)
        self.main_frame.footer = self._get_command_area()


    ### RECEIVE FUNCTIONS ###
    def _initialise_receive_area(self):
        """Initiliases the receive area textbox and filler widgets."""

        self.receive_txt = CustomText('')
        self.receive_filler = CustomFiller(self.receive_txt, 'top')
        self.receive_area = urwid.LineBox(self.receive_filler)

    def set_receive_text(self, text : str, clear : bool=True):
        """Sets the text within the receive textbox.

        Args:
            text (str): the string for the receive textbox
            clear (bool, optional): used to specify if wish to clear current text within the textbox. Defaults to True.
        """

        if clear or self.receive_txt.get_text()[0].strip() == '':
            self.receive_txt.set_text(text)
        else:
            current_text = self.receive_txt.get_text()[0]
            text_rows = self.receive_txt.get_size()[0] + 1 # plus one for new line about to add
            filler_rows = self.receive_filler.get_size()[0]
            if text_rows >= filler_rows:
                current_text = current_text.split('\n')[text_rows-filler_rows:]
                current_text = "\n".join(current_text)
            current_text = current_text + "\n" + text
            self.receive_txt.set_text(current_text)


    ### OPTIONS FUNCTIONS ###
    def _initialise_options_area(self, options : dict):
        """Initiliases the options area widgets and checks options data is valid.

        Args:
            options (dict): the options data dictionary passed to the TerminalUI constructor.

        Raises:
            RuntimeError: if the options is not a dictionary
            RuntimeError: if the value for each key in options is not a dictionary or a tuple
            RuntimeError: if the length of the tuple for the group name key value has a length greater than 3
            RuntimeError: if all options names are not unique
        """     

        # Make sure options is a dictionary
        if type(options) != dict:
            raise RuntimeError('The passed options must be a dictionary')

        # Store options dictionary - need to check all names are unique
        self._options = options
        names = []
        for group_key in self._options:
            # Get data from group name - this is dependent on type
            if type(self._options[group_key]) == dict:
                group_options_dict = self._options[group_key]
            elif type(self._options[group_key]) == tuple:
                # First element in tuple will be a dictionary containing the option names dictionary
                group_options_dict = self._options[group_key][0]
                if len(self._options[group_key]) > 3:
                    raise RuntimeError('The length of the tuple for the value field for the %s key must have a length no greater than 3'%(group_key))
            else:
                raise RuntimeError('The value of the %s key within options should be dictionary or a tuple.'%(group_key))

            # Get keys from group_options_dict
            for option_name in group_options_dict:
                names.append(option_name)

        if len(names) > len(set(names)):
            raise RuntimeError('All option names must be unique.')
        
        # Generate options body - will also result in _option_widgets list been generated
        self._generate_options_body()
        
        # Create listbox and options area
        options_listbox = urwid.ListBox(urwid.SimpleFocusListWalker(self._options_body))
        self.options_area = urwid.LineBox(options_listbox) 
    
    def _generate_options_body(self):
        """Generates the list of widgets for the options area.

        Raises:
            RuntimeError: if it is a selectable option list passed as part of a tuple and the tuple has a length greater than 4
            RuntimeError: if it is an editable value option passed as part of a tuple and the tuple has a length greater than 6
            RuntimeError: if the value for the option name key is not a str, int, float, list or tuple
        """

        self._options_body = [urwid.Text('OPTIONS', align='center'), urwid.Divider()]
        self._option_widgets = []

        for group_key in self._options:
            # Get data from group key
            if type(self._options[group_key]) == dict:
                group_options_data = self._options[group_key]
                show_title = True
                divider_on = True
            elif type(self._options[group_key]) == tuple:
                # First element in tuple will be a dictionary containing the option names dictionary
                group_options_data = self._options[group_key][0]
                if len(self._options[group_key]) == 2:
                    # Assume second element is show title
                    show_title = self._options[group_key][1]
                elif len(self._options[group_key]) == 3:
                    # Assume second element is show title and divider on is third element
                    show_title = self._options[group_key][1]
                    divider_on = self._options[group_key][2] 

            # Add urwid.Text() widget to _options_body to be used as a title if required
            if show_title:
                self._options_body.append(urwid.Text(group_key, align='center'))

            # Add buttons to _options_body 
            for option_name in group_options_data:

                # Create defaults for optional components
                caption = ""
                increment = None
                limits = [None, None]
                enter_fires_change_event = True
                enabled = True

                # Create a UserOption widget
                option_data = group_options_data[option_name]

                # Determine if option_data is a list, a single value or a tuple containing optional UserOption parameters
                if type(option_data) == tuple:
                    # Option data contains optional UserOption parameters

                    # Get value component from option_data tuple
                    value = option_data[0]
                    
                    # Value component will determine if is a list of selectable values or an editable value
                    if type(value) == list:
                        # list of selectable values
                        if len(option_data) == 2:
                            caption = option_data[1]
                        elif len(option_data) == 3:
                            caption = option_data[1]
                            enter_fires_change_event = option_data[2]
                        elif len(option_data) == 4:
                            caption = option_data[1]
                            enter_fires_change_event = option_data[2]
                            enabled = option_data[3]
                        else:
                            raise RuntimeError('The tuple containing the option data when the option is a list of selectable values must have a length less of 4 or less.')
                    elif type(value) == int or type(value) == float or type(value) == str:
                        # editable value
                        if len(option_data) == 2:
                            caption = option_data[1]
                        elif len(option_data) == 3:
                            caption = option_data[1]
                            increment = option_data[2]
                        elif len(option_data) == 4:
                            caption = option_data[1]
                            increment = option_data[2]
                            limits = option_data[3]
                        elif len(option_data) == 5:
                            caption = option_data[1]
                            increment = option_data[2]
                            limits = option_data[3]
                            enter_fires_change_event = option_data[4]
                        elif len(option_data) == 6:
                            caption = option_data[1]
                            increment = option_data[2]
                            limits = option_data[3]
                            enter_fires_change_event = option_data[4]
                            enabled = option_data[5]
                        else:
                            raise RuntimeError('The tuple containing the option data when the option is an editable value must have a length less of 6 or less.')

                    
                elif type(option_data) == list:
                    # option_data is a list of selectable values
                    value = option_data

                elif type(option_data) == int or type(option_data) == float or type(option_data) == str:
                    # option_data is an editable value
                    value = option_data

                else:
                    raise RuntimeError('The option data must be a list, int, float or str, or a tuple where the first element is a list, int, float or str')

                # print(option_name, value, caption, increment, limits, enter_fires_change_event, enabled)

                # Create the actual widget, connect the change signal, append user_option to body and option_widgets list
                user_option = UserOption(option_name, value, caption, increment, limits, enter_fires_change_event, enabled)
                urwid.connect_signal(user_option, 'value_change', self._option_item_selected)
                self._options_body.append(user_option)
                self._option_widgets.append(user_option) # used to find user_options

            # Add urwid.Divider() widget to _options_body if required
            if divider_on:
                self._options_body.append(urwid.Divider()) 

    def _option_item_selected(self, widget : urwid.Widget):
        """Event handler that calls the used defined option_item_selected_callback function passed to the constructor.

        Args:
            widget (urwid.Widget): The widget that raised the change event
        """

        if self._option_item_selected_callback != None:
            self._option_item_selected_callback(self, widget.get_option_name(), widget.get_value(), widget.get_value_index())
        


    def set_option(self, option_name : str, option_value : Union[int, str, float]) -> bool:
        """Used to set an option given an option_name and either a string or integer representing the desired option value. The option string or integer must exist else no change will occur."

        Args:
            option_name (str): The name of the option wish to set
            option_value (Union[str, int]): The string or integer index value for the desired option value

        Returns:
            bool: true if the option was changed, false if no option exists with the passed option_name or if the option string or integer index value for that option does not exist
        """
        
        # Find widget with that name
        widget = None
        for tmp_widget in self._option_widgets:
            if tmp_widget.get_option_name() == option_name:
                widget = tmp_widget
                break
        
        if widget is None:
            return False

        # Try to set option
        try:
            widget.set_value(option_value)
            return True
        except RuntimeError:
            return False


    def get_option(self, option_name : str):
        """Gets the value and index, if the option is a list of selectable values, for the specified option name. The return type for the value will automatically be converted to a int, float or bool if possible, else it will be returned as a string.

        Args:
            option_name (str): the name of the option want the value for

        Returns:
            tuple: (successful, value, index). Successfull will be true if an option with the given name could be found. The value will contain the value of the option. The index will contain the list index if the option is a list of selectable values.
        """

        for widget in self._option_widgets:
            if widget.get_option_name() == option_name:
                return True, widget.get_value(), widget.get_value_index()

        return False, None, None

    def set_option_list(self, option_name : str, options : list, idx : int=0) -> bool:
        """Sets the option list for selectable list option. Will do nothing if the option is an editable value.

        Args:
            option_name (str): the name of the option to set the list for
            options (list): the list of selectable values
            idx (int, optional): the index of the list to display. Defaults to 0.

        Returns:
            bool: if the list was successfull set
        """
        # Find widget with that name
        widget = None
        for widget in self._option_widgets:
            if widget.get_option_name() == option_name:
                return widget.set_option_list(options, idx)
        
        return False

    def enable_option(self, option_name : str, enable : bool):
        """Enables/Disables an option

        Args:
            name (str): The name of the option want to enable/disable
            enable (bool): Specify to true/false if wanting to enable/disable the option
        """

        for widget in self._option_widgets:
            if widget.get_option_name() == option_name:
                widget.enable(enable)

    def option_enter_fires_change_event(self, option_name : str, enter_fires_change_event : bool):
        """Speficies if an option fires a change event when the enter or space key is pressed or if the change event is fired as soon as the option is changed (i.e. left/arrow key alters option).

        Args:
            name (str): The name of the option want to change when event fires
            enable (bool): Specify to true if want to fire change event when enter or space is pressed, or false as soon as option text is changed.
        """

        for widget in self._option_widgets:
            if widget.get_option_name() == option_name:
                widget.enter_fires_change_event(enter_fires_change_event)


    ### TIMER EVENTS ###
    def set_alarm_in(self, sec : float, callback : callable, user_data=None):
        """Set an alarm to fire an event in x seconds and to run a specific function with the passed user data.

        Args:
            sec (float): The number of seconds to wait before firing the event
            callback (callable): The function to call when the event fires
            user_data (optional): Any specified user data. Defaults to None.
        """
        self.main_loop.set_alarm_in(sec, callback, user_data)









import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="terminalui",
    version="0.1.2",
    author="James Mount",
    author_email="jmount1992@gmail.com",
    description="A Terminal User Interface based on URWID with asynchronous input/output that can be easily expanded to your application.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jmount1992/TerminalUI",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'urwid>=2.1.2',
    ]

)